import logging

import time

# from ophyd import EpicsSignal
from epics import PV
from bluesky.utils import CallbackRegistry

from .scan_info import get_scan_info


logger = logging.getLogger(__name__)


class ScanUidMonitor:
    '''Monitor scans via uid PV

    Callback signature looks like:
    * 'start': scan_started(uid)
    * 'stop': scan_finished(uid)

    Parameters
    ----------
    uid_pv : str
        The UID PV name
    '''
    def __init__(self, uid_pv, db):
        self.last_uid = None
        self.cb_registry = CallbackRegistry(allowed_sigs=('start', 'stop'))

        self.cb_registry.connect('start', self.scan_started)
        self.cb_registry.connect('stop', self.scan_finished)

        self.uid_pv = PV(uid_pv, callback=self._uid_changed)
        self.db = db

    def connect(self, sig, func):
        """Register ``func`` to be called when ``sig`` is generated

        Parameters
        ----------
        sig : 'start' or 'stop'
            The signal to monitor
        func : callable
            The function to call

        Returns
        -------
        cid : int
            The callback index. To be used with ``disconnect`` to deregister
            ``func`` so that it will no longer be called when ``sig`` is
            generated
        """
        self.cb_registry.connect(sig, func)

    def disconnect(self, cid):
        """Disconnect the callback registered with callback id *cid*

        Parameters
        ----------
        cid : int
            The callback index and return value from ``connect``
        """
        self.cb_registry.disconnect(cid)

    def _uid_changed(self, value=None, **kwargs):
        '''Uid changed callback'''

        if not value:
            return

        if value != self.last_uid:
            if self.last_uid is not None:
                self._scan_finished(self.last_uid)

            self.last_uid = value
            self._scan_started(value)
        else:
            self._scan_finished(value)

    def _scan_started(self, uid):
        '''Scan started callback, according to uid'''
        logger.debug('Scan started: %s', uid)
        self.cb_registry.process('start', uid)

    def _scan_finished(self, uid):
        '''Scan finished callback, according to uid'''
        logger.debug('Scan finished: %s', uid)
        self.cb_registry.process('stop', uid)

    def scan_started(self, uid, **kwargs):
        '''Default scan started callback

        You can either use the callback registry or for single-use, just
        inherit from this class and override this function.
        '''
        pass

    def scan_finished(self, uid, **kwargs):
        '''Default scan finished callback

        You can either use the callback registry or for single-use, just
        inherit from this class and override this function.
        '''
        pass

    def run():
        print('Monitoring, press Ctrl-C to quit')
        try:
            while True:
                time.sleep(1.)
        except KeyboardInterrupt:
            pass



class ScanBrokerMonitor(ScanUidMonitor):
    '''Monitor scans via uid PV and retrieve metadatastore info

    Callback signature looks like:
    * 'start': scan_started(uid, start=None, **header)
    * 'stop': scan_finished(uid, start=None, stop=None, **header)

    Parameters
    ----------
    uid_pv : str
        The UID PV name
    '''

    def _get_additional_info(self, uid, signal, header):
        '''Additional kwargs for the callbacks'''
        return {}

    def _get_kwargs(self, uid, signal, header):
        '''Add in additional kwargs from _get_additional_info'''
        kwargs = dict(header)
        kwargs.update(self._get_additional_info(uid, signal, header))
        return kwargs

    def _query_db(self, uid):
        '''Query the databroker for a specific uid'''
        return self.db[uid]

    def _scan_started(self, uid):
        logger.debug('Scan started: %s', uid)

        header = self._query_db(uid)
        if header is None:
            return

        if header.get('stop', None) is not None:
            # scan already completed
            logger.debug('scan_started callback but scan is already done')
            self.last_uid = None
            kwargs = self._get_kwargs(uid, 'stop', header)
            self.cb_registry.process('stop', uid, **kwargs)
        else:
            kwargs = self._get_kwargs(uid, 'start', header)
            self.cb_registry.process('start', uid, **kwargs)

    def _scan_finished(self, uid):
        logger.debug('Scan finished: %s', uid)
        header = self._query_db(uid)
        if header is None:
            return
        kwargs = self._get_kwargs(uid, 'stop', header)
        self.cb_registry.process('stop', uid, **kwargs)

    def scan_started(self, uid, start=None, **header):
        pass

    def scan_finished(self, uid, start=None, stop=None, **header):
        pass


class HxnScanMonitor(ScanBrokerMonitor):
    '''Monitor scans via uid PV, retrieve mds info, and get hxn-specific info

    Callback signature looks like:
    * 'start': scan_started(uid, start={}, hxn_info={}, **header)
    * 'stop': scan_finished(uid, start={}, stop={}, hxn_info={}, **header)

    Parameters
    ----------
    uid_pv : str
        The UID PV name
    '''

    def _get_additional_info(self, uid, signal, header):
        return {'hxn_info': get_scan_info(header)}

    def scan_started(self, uid, start=None, hxn_info=None,
                     **header):
        pass

    def scan_finished(self, uid, start=None, stop=None, hxn_info=None,
                      **header):
        pass


def _test(db):
    import sys

    try:
        uid_pv = sys.argv[1]
    except IndexError:
        uid_pv = 'XF:03IDC-ES{BS-Scan}UID-I'

    print('Using uid pv: {}'.format(uid_pv))
    # uid_mon = ScanUidMonitor(uid_pv)
    # broker_mon = ScanBrokerMonitor(uid_pv)
    hxn_mon = HxnScanMonitor(uid_pv, db)

    hxn_mon.run()


if __name__ == '__main__':
    from databroker.databroker import DataBroker
    logging.basicConfig(level=logging.DEBUG)
    _test(DataBroker)



from collections import OrderedDict, namedtuple
import uuid
import itertools
import logging
import time
import numpy as np

from ophyd import (Component as Cpt, Signal)
from ophyd.areadetector.plugins import PluginBase
from ophyd.status import DeviceStatus
from ophyd.device import (BlueskyInterface, Staged)

from nslsii.detectors.xspress3 import (XspressTrigger, Xspress3Detector,
                                       Xspress3FileStore, Xspress3ROI)
from .trigger_mixins import HxnModalBase

import pandas as pd

logger = logging.getLogger(__name__)


class HxnXspressTrigger(HxnModalBase, BlueskyInterface):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status = None
        self._acquisition_signal = self.settings.acquire
        self._abs_trigger_count = 0

    def unstage(self):
        ret = super().unstage()
        try:
            self._acquisition_signal.clear_sub(self._acquire_changed)
        except KeyError:
            pass

        return ret

    def _acquire_changed(self, value=None, old_value=None, **kwargs):
        "This is called when the 'acquire' signal changes."
        if self._status is None:
            return
        if (old_value == 1) and (value == 0):
            # Negative-going edge means an acquisition just finished.
            self._status._finished()

    def mode_internal(self):
        self._abs_trigger_count = 0
        self.stage_sigs[self.external_trig] = False
        self.stage_sigs[self.settings.acquire_time] = self.count_time.get()
        self._acquisition_signal.subscribe(self._acquire_changed)

    def mode_external(self):
        ms = self.mode_settings
        # NOTE: these are used in Xspress3Filestore.stage
        self.stage_sigs[self.external_trig] = True
        self.stage_sigs[self.total_points] = ms.total_points.get()
        self.stage_sigs[self.spectra_per_point] = 1
        self.stage_sigs[self.settings.acquire_time] = 0.001

        # NOTE: these are taken care of in Xspress3Filestore
        # self.stage_sigs[self.settings.trigger_mode] = 'TTL Veto Only'
        # self.stage_sigs[self.settings.num_images] = total_capture

    def _dispatch_channels(self, trigger_time):
        self._abs_trigger_count += 1
        channels = self._channels.values()
        for sn in self.read_attrs:
            ch = getattr(self, sn)
            if ch in channels:
                self.dispatch(ch.name, trigger_time)

    def trigger_internal(self):
        if self._staged != Staged.yes:
            raise RuntimeError("not staged")
        self._status = DeviceStatus(self)
        self.settings.erase.put(1)
        self._acquisition_signal.put(1, wait=False)
        self._dispatch_channels(trigger_time=time.time())
        return self._status

    def trigger_external(self):
        if self._staged != Staged.yes:
            raise RuntimeError("not staged")

        self._status = DeviceStatus(self)
        self._status._finished()
        if self.mode_settings.scan_type.get() != 'fly':
            self._dispatch_channels(trigger_time=time.time())
            # fly-scans take care of dispatching on their own

        return self._status

    def stage(self):
        staged = super().stage()
        mode = self.mode_settings.mode.get()
        if mode == 'external':
            # In external triggering mode, the devices is only triggered once
            # at stage
            self.settings.erase.put(1, wait=True)
            self._acquisition_signal.put(1, wait=False)
        return staged


class HxnXspress3DetectorBase(HxnXspressTrigger, Xspress3Detector):
    roi_data = Cpt(PluginBase, 'ROIDATA:')
    flyer_timestamps = Cpt(Signal)

    @property
    def hdf5_filename(self):
        return self.hdf5._fn

    def describe_collect(self):
        desc = Xspress3FileStore.describe(self.hdf5)

        for roi in self.enabled_rois:
            desc[roi.name] = dict(source='FILE:TBD', shape=[], dtype='number')

        return [desc]

    def bulk_read(self, timestamps=None):
        # TODO not compatible with collect() just yet due to the values
        #      returned
        if timestamps is None:
            timestamps = self.flyer_timestamps.get()

        if timestamps is None:
            raise ValueError('Timestamps must be set first')

        channels = self.channels
        count = len(timestamps)
        if count == 0:
            return {}

        ch_uids = {}
        for ch in channels:
            ch_uids[ch] = self.hdf5._reg.bulk_register_datum_table(
                self.hdf5._filestore_res, pd.DataFrame({'frame': range(count),
                                                        'channel': ch}))

        return OrderedDict((self.hdf5.mds_keys[ch], ch_uids[ch])
                           for ch in channels)

    def fly_collect_rois(self, rois=None, *, ignore_get_failures=True):
        '''Read ROI data from the PVs

        Parameters
        ----------
        rois : sequence of Xspress3ROI instances, optional
            If unspecified, uses all currently enabled ROIs
        ignore_get_failures : bool, optional
            Ignore pyepics-related failures - will
        '''
        if rois is None:
            rois = self.enabled_rois

        num_points = self.settings.num_images.get()
        RoiTuple = namedtuple('Xspress3ROITuple',
                              ['bin_low', 'bin_high',
                               'ev_low', 'ev_high',
                               'value', 'value_sum',
                               'enable'])

        for roi in self.enabled_rois:
            try:
                roi_data = roi.settings.array_data.get(count=num_points,
                                                       use_monitor=False)
            except Exception as ex:
                logger.error('Failed to get ROI data', exc_info=ex)
                if not ignore_get_failures:
                    raise
                roi_data = np.zeros(num_points)

            try:
                roi_data = roi_data[:num_points]
            except TypeError as ex:
                logger.error('Failed to get ROI data', exc_info=ex)
                if not ignore_get_failures:
                    raise
                roi_data = np.zeros(num_points)

            roi_info = RoiTuple(bin_low=roi.bin_low.get(),
                                bin_high=roi.bin_high.get(),
                                ev_low=roi.ev_low.get(),
                                ev_high=roi.ev_high.get(),
                                value=roi_data,
                                value_sum=None,
                                enable=roi.enable.get())

            yield roi.name, roi_info

    def stop(self, success=False):
        super().stop(success=success)

        logger.info('Ensuring detector %r capture stopped...',
                    self.name)
        self.settings.acquire.set(0).wait()
        self.hdf5.stop(success=success)
        logger.info('... detector %r ok', self.name)



from __future__ import absolute_import

# imports to expose out to world
from .timepix import TimepixHDF5Handler


def register(db):
    from .xspress3 import register
    register(db)
    from .timepix import register
    register(db)



import collections
from ophyd import (EpicsScaler, Device,
                   Component as Cpt, DynamicDeviceComponent as DDC,
                   EpicsSignal, Signal)
from ophyd.mca import EpicsMCARecord
from ophyd.status import DeviceStatus
from ophyd.device import Staged
from .detectors.trigger_mixins import HxnModalBase


class MinimalCalcRecord(Device):
    value = Cpt(EpicsSignal, '.VAL')
    equation = Cpt(EpicsSignal, '.CALC$', string=True)


def _scaler_calc_records(suffix_format='', range_=None):
    if range_ is None:
        range_ = range(1, 9)

    recs = collections.OrderedDict()

    for i in range_:
        attr = 'calc{}'.format(i)
        recs[attr] = (MinimalCalcRecord, suffix_format.format(i), {})

    return recs


class StruckMCA(EpicsMCARecord):
    def __init__(self, prefix, *, index=None, **kwargs):
        self.index = index
        super().__init__(prefix, **kwargs)

    def stop(self, success=False):
        pass


def _struck_mca_records(prefix_format, range_):
    recs = collections.OrderedDict()

    for i in range_:
        attr = 'mca{:02d}'.format(i)
        mca_prefix = prefix_format.format(i)
        recs[attr] = (StruckMCA, mca_prefix, dict(index=i))

    recs['mca_attrs'] = (Signal, None, dict(value=tuple(recs.keys())))
    return recs


class EpicsScalerWithCalc(EpicsScaler):
    calculations = DDC(_scaler_calc_records('_calc{}', range(1, 9)))
    enable_calculations = Cpt(EpicsSignal, '_calcEnable')


class StruckScaler(EpicsScalerWithCalc):
    mca_count = 32

    mcas = DDC(_struck_mca_records('Mca:{}', range(1, mca_count + 1)))

    acquire_mode = Cpt(EpicsSignal, 'AcquireMode')
    acquiring = Cpt(EpicsSignal, 'Acquiring')
    asyn = Cpt(EpicsSignal, 'Asyn')
    channel1_source = Cpt(EpicsSignal, 'Channel1Source')
    channel_advance = Cpt(EpicsSignal, 'ChannelAdvance', string=True)
    client_wait = Cpt(EpicsSignal, 'ClientWait')
    count_on_start = Cpt(EpicsSignal, 'CountOnStart')
    current_channel = Cpt(EpicsSignal, 'CurrentChannel')
    disable_auto_count = Cpt(EpicsSignal, 'DisableAutoCount')
    do_read_all = Cpt(EpicsSignal, 'DoReadAll')
    dwell = Cpt(EpicsSignal, 'Dwell')
    elapsed_real = Cpt(EpicsSignal, 'ElapsedReal')
    enable_client_wait = Cpt(EpicsSignal, 'EnableClientWait')
    erase_all = Cpt(EpicsSignal, 'EraseAll')
    erase_start = Cpt(EpicsSignal, 'EraseStart')
    firmware = Cpt(EpicsSignal, 'Firmware')
    hardware_acquiring = Cpt(EpicsSignal, 'HardwareAcquiring')
    input_mode = Cpt(EpicsSignal, 'InputMode')
    max_channels = Cpt(EpicsSignal, 'MaxChannels')
    model = Cpt(EpicsSignal, 'Model')
    mux_output = Cpt(EpicsSignal, 'MUXOutput')
    nuse_all = Cpt(EpicsSignal, 'NuseAll')
    output_mode = Cpt(EpicsSignal, 'OutputMode')
    output_polarity = Cpt(EpicsSignal, 'OutputPolarity')
    prescale = Cpt(EpicsSignal, 'Prescale')
    preset_real = Cpt(EpicsSignal, 'PresetReal')
    read_all = Cpt(EpicsSignal, 'ReadAll')
    read_all_once = Cpt(EpicsSignal, 'ReadAllOnce')
    set_acquiring = Cpt(EpicsSignal, 'SetAcquiring')
    set_client_wait = Cpt(EpicsSignal, 'SetClientWait')
    snl_connected = Cpt(EpicsSignal, 'SNL_Connected')
    software_channel_advance = Cpt(EpicsSignal, 'SoftwareChannelAdvance')
    start_all = Cpt(EpicsSignal, 'StartAll')
    stop_all = Cpt(EpicsSignal, 'StopAll')
    user_led = Cpt(EpicsSignal, 'UserLED')
    wfrm = Cpt(EpicsSignal, 'Wfrm')

    def __init__(self, prefix, **kwargs):
        super().__init__(prefix, **kwargs)

        mca_attrs = self.mcas.mca_attrs.get()
        self.mca_by_index = {getattr(self.mcas, attr).index: getattr(self.mcas,
                                                                     attr)
                             for attr in mca_attrs}


class HxnTriggeringScaler(HxnModalBase, StruckScaler):
    def __init__(self, prefix, *, scan_type_triggers=None, **kwargs):
        super().__init__(prefix, **kwargs)

        if scan_type_triggers is None:
            scan_type_triggers = {'step': [],
                                  'fly': [],
                                  }

        self.scan_type_triggers = dict(scan_type_triggers)

        # Scaler 1 should be in output mode 1 to properly trigger
        self.stage_sigs[self.output_mode] = 'Mode 1'
        # Ensure that the scaler isn't counting in mcs mode for any reason
        self.stage_sigs[self.stop_all] = 1
        # Ensure that the scaler isn't counting
        self.stage_sigs[self.count] = 0

        self.external_signals = [self.erase_start,
                                 self.start_all,
                                 self.input_mode,
                                 self.channel_advance,
                                 self.nuse_all,
                                 self.preset_real,
                                 self.dwell]

    def mode_internal(self):
        super().mode_internal()

        settings = self.mode_settings

        # ensure that no external signals are in the stage sigs list:
        for ext_sig in self.external_signals:
            if ext_sig in self.stage_sigs:
                del self.stage_sigs[ext_sig]

        scan_type = settings.scan_type.get()

        triggers = self.scan_type_triggers[scan_type]
        settings.triggers.put(list(triggers))

        self.stop_all.put(1, wait=True)
        self.count.put(0, wait=True)
        self.stage_sigs[self.preset_time] = self.count_time.get()

    def mode_external(self):
        super().mode_external()

        if self.preset_time in self.stage_sigs:
            del self.stage_sigs[self.preset_time]

        self.stage_sigs[self.preset_real] = 0.0
        self.stage_sigs[self.dwell] = 0.0
        self.stage_sigs[self.channel_advance] = 'External'
        self.stage_sigs[self.input_mode] = 'Mode 2'
        self.stage_sigs[self.nuse_all] = self.mode_settings.total_points.get()

        self.stage_sigs[self.erase_start] = 1
        self.stage_sigs.move_to_end(self.erase_start)

    def trigger(self):
        if self._staged != Staged.yes:
            raise RuntimeError("This detector is not ready to trigger."
                               "Call the stage() method before triggering.")

        mode = self.mode_settings.mode.get()
        mode_trigger = getattr(self, 'trigger_{}'.format(mode))
        return mode_trigger()

    def trigger_internal(self):
        return super().trigger()

    def trigger_external(self):
        self._status = DeviceStatus(self)
        self._status._finished()
        return self._status


class HxnScaler(StruckScaler):
    pass



import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d, interp2d

from hxntools.scan_info import ScanInfo


def fly2d_grid(hdr, x_data=None, y_data=None, plot=False):
    '''Get ideal gridded points for a 2D flyscan'''
    hdr = hdr['start']
    try:
        nx, ny = hdr['dimensions']
    except ValueError:
        raise ValueError('Not a 2D flyscan (dimensions={})'
                         ''.format(hdr['dimensions']))

    rangex, rangey = hdr['scan_range']
    width = rangex[1] - rangex[0]
    height = rangey[1] - rangey[0]

    macros = eval(hdr['subscan_0']['macros'], dict(array=np.array))
    start_x, start_y = macros['scan_starts']
    dx = width / nx
    dy = height / ny
    grid_x = np.linspace(start_x, start_x + width + dx / 2, nx)
    grid_y = np.linspace(start_y, start_y + height + dy / 2, ny)

    if plot:
        mesh_x, mesh_y = np.meshgrid(grid_x, grid_y)
        plt.figure()
        if x_data is not None and y_data is not None:
            plt.scatter(x_data, y_data, c='blue', label='actual')
        plt.scatter(mesh_x, mesh_y, c='red', label='gridded',
                    alpha=0.5)
        plt.legend()
        plt.show()

    return grid_x, grid_y


def interp2d_scan(hdr, x_data, y_data, spectrum, *, kind='linear',
                  plot_points=False, **kwargs):
    '''Interpolate a 2D flyscan over a grid'''
    new_x, new_y = fly2d_grid(hdr, x_data, y_data, plot=plot_points)

    f = interp2d(x_data, y_data, spectrum, kind=kind, **kwargs)
    return f(new_x, new_y)


def interp1d_scan(hdr, x_data, y_data, spectrum, kind='linear',
                  plot_points=False, **kwargs):
    '''Interpolate a 2D flyscan only over the fast-scanning direction'''
    grid_x, grid_y = fly2d_grid(hdr, x_data, y_data, plot=plot_points)
    x_data = fly2d_reshape(hdr, x_data, verbose=False)

    spectrum2 = np.zeros_like(spectrum)
    for row in range(len(grid_y)):
        spectrum2[row, :] = interp1d(x_data[row, :], spectrum[row, :],
                                     kind=kind, bounds_error=False,
                                     **kwargs)(grid_x)

    return spectrum2


def fly2d_reshape(hdr, spectrum, verbose=True, copy=False):
    '''Reshape a 1D array to match the shape of a 2D flyscan'''
    info = ScanInfo(hdr)
    start_doc = hdr['start']

    try:
        nx, ny = info.dimensions
    except Exception:
        raise ValueError('Not a 2D flyscan (dimensions={})'
                         ''.format(info.dimensions))

    num = nx * ny
    spectrum = np.asarray(spectrum).flatten()

    if len(spectrum) < num:
        spectrum_ = np.ones(num) * np.min(spectrum)
        spectrum_[:len(spectrum)] = spectrum.flatten()
        spectrum = spectrum_
    elif copy:
        spectrum = spectrum.copy()

    try:
        spectrum = spectrum.reshape((ny, nx))
    except Exception as ex:
        if verbose:
            print('\tUnable to reshape data to (%d, %d) (%s: %s)'
                  '' % (nx, ny, ex.__class__.__name__, ex))
    else:
        fly_type = start_doc.get('fly_type', None)
        if fly_type in ('pyramid', ):
            # Pyramid scans' odd rows are flipped:
            subscan_dims = start_doc.get('subscan_dims', None)
            if subscan_dims is None:
                spectrum[1::2, :] = spectrum[1::2, ::-1]
            else:
                i = 0
                for nx, ny in subscan_dims:
                    start = i + 1
                    end = i + ny
                    spectrum[start:end:2, :] = spectrum[start:end:2, ::-1]
                    i += ny

        return spectrum



from __future__ import print_function
import epics
import time


def shutter_open():
    open_pv = epics.PV('XF:03IDB-PPS{PSh}Cmd:Opn-Cmd')
    closed_pv = epics.PV('XF:03IDB-PPS{PSh}Sts:Cls-Sts')

    if closed_pv.get() != 0:
        print('Opening the photon shutter...')
        open_pv.put(1, wait=True)
        while closed_pv.get() == 0:
            print('Waiting for the shutter to open...')
            time.sleep(1.0)


def shutter_close():
    print('Closing the photon shutter...')
    close_pv = epics.PV('XF:03IDB-PPS{PSh}Cmd:Cls-Cmd')
    close_pv.put(1)



import asyncio
import functools
import logging

from cycler import cycler
from boltons.iterutils import chunked

from bluesky import (plans, Msg)
import bluesky.preprocessors as bpp
import bluesky.plan_stubs as bps
from bluesky import plan_patterns

from ophyd import (Device, Component as Cpt, EpicsSignal)
from .detectors.trigger_mixins import HxnModalBase

logger = logging.getLogger(__name__)


class ScanID(Device):
    next_scan_id_proc = Cpt(EpicsSignal, 'NextScanID-Cmd.PROC')
    scan_id = Cpt(EpicsSignal, 'ScanID-I')

    def get_next_scan_id(self):
        last_id = int(self.scan_id.get())
        self.next_scan_id_proc.put(1, wait=True)

        new_id = int(self.scan_id.get())
        if last_id == new_id:
            raise RuntimeError('Scan ID unchanged. Check hxnutil IOC.')
        return new_id


dev_scan_id = ScanID('XF:03IDC-ES{Status}', name='dev_scan_id')


def get_next_scan_id():
    dev_scan_id.wait_for_connection()
    return dev_scan_id.get_next_scan_id()


def one_nd_step(detectors, step, pos_cache):
    """
    Inner loop of an N-dimensional step scan

    This is the default function for ``per_step`` param`` in ND plans.

    Parameters
    ----------
    detectors : iterable
        devices to read
    step : dict
        mapping motors to positions in this step
    pos_cache : dict
        mapping motors to their last-set positions
    """
    def move():
        yield Msg('checkpoint')
        for motor, pos in step.items():
            if pos == pos_cache[motor]:
                # This step does not move this motor.
                continue
            yield from bps.mov(motor, pos)
            pos_cache[motor] = pos

    motors = step.keys()
    yield from move()
    yield from bps.trigger_and_read(list(detectors) + list(motors))



@asyncio.coroutine
def cmd_scan_setup(msg):
    detectors = msg.kwargs['detectors']
    total_points = msg.kwargs['total_points']
    count_time = msg.kwargs['count_time']

    modal_dets = [det for det in detectors
                  if isinstance(det, HxnModalBase)]

    mode = 'internal'
    for det in modal_dets:
        logger.debug('[internal trigger] Setting up detector %s', det.name)
        settings = det.mode_settings

        # Ensure count time is set prior to mode setup
        det.count_time.put(count_time)

        # start by using internal triggering
        settings.mode.put(mode)
        settings.scan_type.put('step')
        settings.total_points.put(total_points)
        det.mode_setup(mode)

    # the mode setup above should update to inform us which detectors
    # are externally triggered, in the form of the list in
    #   mode_settings.triggers
    # so update each of those to use external triggering
    triggered_dets = [det.mode_settings.triggers.get()
                      for det in modal_dets]
    triggered_dets = [triggers for triggers in triggered_dets
                      if triggers is not None]
    triggered_dets = set(sum(triggered_dets, []))

    logger.debug('These detectors will be externally triggered: %s',
                 ', '.join(det.name for det in triggered_dets))

    mode = 'external'
    for det in triggered_dets:
        logger.debug('[external trigger] Setting up detector %s', det)
        det.mode_settings.mode.put(mode)
        det.mode_setup(mode)


def setup(*, RE, debug_mode=False):
    @asyncio.coroutine
    def cmd_next_scan_id(msg):
        RE.md['scan_id'] = get_next_scan_id() - 1

    @asyncio.coroutine
    def _debug_next_scan_id(cmd):
        print('debug_next_scan_id')
        RE.md['scan_id'] = 0

    RE.register_command('hxn_scan_setup', cmd_scan_setup)

    if debug_mode:
        RE.register_command('hxn_next_scan_id', _debug_next_scan_id)
    else:
        RE.register_command('hxn_next_scan_id', cmd_next_scan_id)


def _pre_scan(dets, total_points, count_time):
    yield Msg('hxn_next_scan_id')
    yield Msg('hxn_scan_setup', detectors=dets, total_points=total_points,
              count_time=count_time)


@functools.wraps(plans.count)
def count(dets, num=1, delay=None, time=None, *, md=None):
    yield from _pre_scan(dets, total_points=num, count_time=time)
    return (yield from plans.count(dets, num=num, delay=delay, md=md))


@functools.wraps(plans.scan)
def absolute_scan(dets, motor, start, finish, intervals, time=None, *,
                  md=None):
    yield from _pre_scan(dets, total_points=intervals + 1, count_time=time)
    return (yield from plans.scan(dets, motor, start, finish, intervals, md=md))


@functools.wraps(plans.relative_scan)
def relative_scan(dets, motor, start, finish, intervals, time=None, *,
                  md=None):
    yield from _pre_scan(dets, total_points=intervals + 1, count_time=time)
    return (yield from plans.relative_scan(dets, motor, start, finish,
                                           intervals+1, md=md))


@functools.wraps(plans.spiral_fermat)
def absolute_fermat(dets, x_motor, y_motor, x_start, y_start, x_range,
                    y_range, dr, factor, time=None, *, per_step=None,
                    md=None, tilt=0.0):
    cyc = plan_patterns.spiral_fermat(x_motor, y_motor, x_motor.position,
                                      y_motor.position, x_range, y_range, dr,
                                      factor, tilt=tilt)
    total_points = len(cyc)

    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.spiral_fermat(dets, x_motor, y_motor, x_start,
                                           y_start, x_range, y_range, dr, factor,
                                per_step=per_step, md=md, tilt=tilt))


@functools.wraps(plans.relative_spiral_fermat)
def relative_fermat(dets, x_motor, y_motor, x_range, y_range, dr,
                    factor, time=None, *, per_step=None, md=None, tilt=0.0):
    cyc = plan_patterns.spiral_fermat(x_motor, y_motor, x_motor.position,
                                      y_motor.position, x_range, y_range, dr,
                                      factor, tilt=tilt)
    total_points = len(cyc)

    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.relative_spiral_fermat(
        dets,
        x_motor, y_motor, x_range, y_range, dr, factor,
        per_step=per_step, md=md, tilt=tilt))


@functools.wraps(plans.spiral)
def absolute_spiral(dets, x_motor, y_motor, x_start, y_start, x_range,
                    y_range, dr, nth, time=None, *, per_step=None,
                    md=None, tilt=0.0):
    cyc = plan_patterns.spiral_simple(x_motor, y_motor, x_motor.position,
                                      y_motor.position, x_range, y_range, dr,
                                      nth, tilt=tilt)
    total_points = len(cyc)

    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.spiral(
        dets, x_motor, y_motor, x_start, y_start,
        x_range, y_range, dr, nth, per_step=per_step,
        md=md, tilt=tilt))


@functools.wraps(plans.relative_spiral)
def relative_spiral(dets, x_motor, y_motor, x_range, y_range, dr, nth,
                    time=None, *, per_step=None, md=None, tilt=0.0):
    cyc = plan_patterns.spiral_simple(x_motor, y_motor, x_motor.position,
                                      y_motor.position, x_range, y_range, dr,
                                      nth, tilt=tilt)
    total_points = len(cyc)

    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.relative_spiral(
        dets, x_motor, y_motor, x_range,
        y_range, dr, nth, per_step=per_step,
        md=md, tilt=tilt))


@functools.wraps(plans.outer_product_scan)
def absolute_mesh(dets, *args, time=None, md=None):
    if (len(args) % 4) == 1:
        if time is not None:
            raise ValueError('wrong number of positional arguments')
        args, time = args[:-1], args[-1]

    total_points = 1
    new_args = []
    add_snake = False
    for motor, start, stop, num in chunked(args, 4):
        total_points *= num
        new_args += [motor, start, stop, num]
        if add_snake:
            new_args += [False]
        add_snake = True

    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.outer_product_scan(dets, *new_args, md=md,
                                                per_step=one_nd_step))


@functools.wraps(absolute_mesh)
def relative_mesh(dets, *args, time=None, md=None):
    plan = absolute_mesh(dets, *args, time=time, md=md)
    plan = bpp.relative_set_wrapper(plan)  # re-write trajectory as relative
    return (yield from bpp.reset_positions_wrapper(plan))


def _get_a2_args(*args, time=None):
    if (len(args) % 3) == 2:
        if time is not None:
            raise ValueError('Wrong number of positional arguments')
        args, time = args[:-1], args[-1]

    return args, time


@functools.wraps(plans.inner_product_scan)
def a2scan(dets, *args, time=None, md=None):
    args, time = _get_a2_args(*args, time=time)
    total_points = int(args[0])
    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.inner_product_scan(dets, *args, md=md,
                                                per_step=one_nd_step))


@functools.wraps(plans.relative_inner_product_scan)
def d2scan(dets, *args, time=None, md=None):
    args, time = _get_a2_args(*args, time=time)
    total_points = int(args[0])
    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.relative_inner_product_scan(dets, *args, md=md,
                                                         per_step=one_nd_step))


def scan_steps(dets, *args, time=None, per_step=None, md=None):
    '''
    Absolute scan over an arbitrary N-dimensional trajectory.

    Parameters
    ----------
    ``*args`` : {Positioner, list/sequence}
        Patterned like
            (``motor1, motor1_positions, ..., motorN, motorN_positions``)
        Where motorN_positions is a list/tuple/sequence of absolute positions
        for motorN to go to.
    time : float, optional
        applied to any detectors that have a `count_time` setting
    per_step : callable, optional
        hook for cutomizing action of inner loop (messages per step)
        See docstring of bluesky.plans.one_nd_step (the default) for
        details.
    md : dict, optional
        metadata
    '''
    if len(args) % 2 == 1:
        if time is not None:
            raise ValueError('Wrong number of positional arguments')
        args, time = args[:-1], args[-1]

    cyclers = [cycler(motor, steps) for motor, steps in chunked(args, 2)]
    cyc = sum(cyclers[1:], cyclers[0])
    total_points = len(cyc)

    if md is None:
        md = {}

    from collections import ChainMap

    md = ChainMap(md, {'plan_name': 'scan_steps'})

    plan = plans.scan_nd(dets, cyc, md=md, per_step=per_step)
    plan = plans.configure_count_time_wrapper(plan, time)

    yield from _pre_scan(dets, total_points=total_points, count_time=time)
    return (yield from plans.reset_positions_wrapper(plan))



import logging
import collections
import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


def _eval(scan_args):
    '''Evaluate scan arguments, replacing OphydObjects with NamedObjects'''

    class NamedObject:
        def __init__(self, name):
            self.name = name

    def no_op():
        def no_op_inner(*args, name=None, **kwargs):
            if name is not None:
                return NamedObject(name)

        return no_op_inner

    return eval(scan_args, collections.defaultdict(no_op))


scan_types = dict(
    v0=dict(step_1d=('InnerProductAbsScan', 'HxnInnerAbsScan',
                     'InnerProductDeltaScan', 'HxnInnerDeltaScan', 'AbsScan',
                     'HxnAbsScan', 'DeltaScan', 'HxnDeltaScan'),
            step_2d=('OuterProductAbsScan', 'HxnOuterAbsScan', 'relative_mesh',
                     'absolute_mesh'),
            spiral=('HxnFermatPlan', 'relative_fermat', 'absolute_fermat',
                    'relative_spiral', 'absolute_spiral'),
            fly=('FlyPlan1D', 'FlyPlan2D'),
            ),
    v1=dict(fly=('FlyPlan1D', 'FlyPlan2D', 'FlyStep1D'),
            # everything else can be done with plan_patterns
            ),
)


def _get_scan_info_bs_v0(header):
    info = {'num': 0,
            'dimensions': [],
            'motors': [],
            'range': [],
            'pyramid': False,
            }

    start_doc = header['start']
    try:
        scan_args = start_doc['scan_args']
    except KeyError:
        try:
            scan_args = start_doc['plan_args']
        except KeyError:
            logger.error('No scan args for scan %s', start_doc['uid'])
            return info

    try:
        scan_type = start_doc['scan_type']
    except KeyError:
        try:
            scan_type = start_doc['plan_type']
        except KeyError:
            logger.error('No plan type for scan %s', start_doc['uid'])
            return info

    motors = None
    range_ = None
    pyramid = False
    motor_keys = None
    exposure_time = 0.0
    dimensions = []

    scan_type_info = scan_types['v0']
    step_1d = scan_type_info['step_1d']
    step_2d = scan_type_info['step_2d']
    spiral_scans = scan_type_info['spiral']
    fly_scans = scan_type_info['fly']

    if scan_type in fly_scans:
        dimensions = start_doc['dimensions']
        try:
            motors = start_doc['motors']
        except KeyError:
            motors = start_doc['axes']

        pyramid = start_doc['fly_type'] == 'pyramid'
        exposure_time = float(scan_args.get('exposure_time', 0.0))

        logger.debug('Scan %s (%s) is a fly-scan (%s) of axes %s '
                     'with per-frame exposure time of %.3f s',
                     start_doc.scan_id, start_doc.uid, scan_type,
                     motors, exposure_time)
        try:
            range_ = start_doc['scan_range']
        except KeyError:
            try:
                range_ = [(float(start_doc['scan_start']),
                           float(start_doc['scan_end']))]
            except (KeyError, ValueError):
                pass
    elif scan_type in step_2d:
        logger.debug('Scan %s (%s) is an ND scan (%s)', start_doc.scan_id,
                     start_doc.uid, scan_type)

        try:
            args = _eval(scan_args['args'])
        except Exception:
            pass

        # 2D mesh scan
        try:
            motors = [arg.name for arg in args[::5]]
        except Exception:
            motors = []

        exposure_time = float(start_doc.get('exposure_time', 0.0))
        try:
            dimensions = args[3::5]
            range0 = args[1::5]
            range1 = args[2::5]
            range_ = list(zip(range0, range1))
        except Exception:
            dimensions = []
            range_ = []

    elif scan_type in spiral_scans:
        motor_keys = ['x_motor', 'y_motor']
        dimensions = [int(start_doc['num'])]
        exposure_time = float(scan_args.get('exposure_time', 0.0))
        logger.debug('Scan %s (%s) is a fermat scan (%s) %d points, '
                     'with per-point exposure time of %.3f s',
                     start_doc.scan_id, start_doc.uid, scan_type,
                     dimensions[0], exposure_time)
        try:
            range_ = [(float(start_doc['x_range']),
                       float(start_doc['y_range']))]
        except (KeyError, ValueError):
            pass

    elif scan_type in step_1d or 'num' in start_doc:
        logger.debug('Scan %s (%s) is a 1D scan (%s)', start_doc.scan_id,
                     start_doc.uid, scan_type)
        exposure_time = float(start_doc.get('exposure_time', 0.0))
        # 1D scans
        try:
            dimensions = [int(start_doc['num'])]
        except KeyError:
            # some scans with the bluesky md changes didn't save num
            dimensions = []
        motor_keys = ['motor']
    else:
        msg = 'Unrecognized scan type (uid={} {})'.format(start_doc.uid,
                                                          scan_type)
        raise RuntimeError(msg)

    if motor_keys:
        motors = []
        for key in motor_keys:
            try:
                motors.append(_eval(start_doc[key]).name)
            except Exception:
                pass

    num = np.product(dimensions)

    info['num'] = num
    info['dimensions'] = dimensions
    info['motors'] = motors
    info['range'] = range_
    info['pyramid'] = pyramid
    info['exposure_time'] = exposure_time
    return info


def get_scan_info(header):
    start_doc = header['start']
    if 'scan_args' in start_doc:
        return _get_scan_info_bs_v0(header)
    elif 'plan_args' in start_doc:
        return _get_scan_info_bs_v1(header)
    else:
        raise RuntimeError('Unknown start document information')


def _get_scan_info_bs_v1(header):
    start_doc = header['start']
    info = {'num': 0,
            'dimensions': [],
            'motors': [],
            'range': {},
            'pyramid': False,
            }

    plan_type = start_doc['plan_type']
    plan_name = start_doc['plan_name']

    range_ = None
    pyramid = False

    plan_type_info = scan_types['v1']
    fly_scans = plan_type_info['fly']

    motors = start_doc['motors']

    if plan_type in fly_scans:
        logger.debug('Scan %s (%s) is a fly scan (%s %s)', start_doc.scan_id,
                     start_doc.uid, plan_type, plan_name)
        shape = start_doc['shape']
        pyramid = start_doc['fly_type'] == 'pyramid'
        range_ = dict(zip(motors, start_doc['scan_range']))
    else:
        try:
            pattern_module = start_doc['plan_pattern_module']
            pattern = start_doc['plan_pattern']
            pattern_args = start_doc['plan_pattern_args']
        except KeyError as ex:
            msg = ('Unrecognized plan type/name (uid={} name={} type={}; '
                   'missing key: {!r})'.format(start_doc.uid, plan_name,
                                               plan_type, ex))
            raise RuntimeError(msg)

        module = __import__(pattern_module, fromlist=[''])
        pattern_fcn = getattr(module, pattern)
        logger.debug('Calling plan pattern function: %s with kwargs %s',
                     pattern_fcn, pattern_args)
        points = pattern_fcn(**pattern_args)

        if isinstance(points, (list, np.ndarray)):
            num = len(points)
            range_ = {motors[0]: (np.min(points), np.max(points))}
        else:
            cyc = points
            num = len(cyc)
            try:
                # cycler >= 0.10
                positions = cyc.by_key()
            except AttributeError:
                positions = collections.defaultdict(lambda: [])
                for pt in iter(cyc):
                    for mtr, pos in pt.items():
                        positions[mtr].append(pos)

            range_ = {_eval(mtr).name: (np.min(positions[mtr]),
                                        np.max(positions[mtr]))
                      for mtr in cyc.keys}

        shape = start_doc.get('shape', None)
        if shape is None:
            shape = [num]

    info['num'] = np.product(shape)
    info['dimensions'] = shape
    info['motors'] = motors
    info['range'] = range_
    info['pyramid'] = pyramid
    return info


class ScanInfo(object):
    def __init__(self, header, stream='primary'):
        self.header = header
        self.start_doc = header['start']
        self.descriptors = header['descriptors']
        self.key = None

        if 'scan_args' in self.start_doc and stream == 'primary':
            stream = None

        self.stream = stream
        for key, value in get_scan_info(self.header).items():
            logger.debug('Scan info %s=%s', key, value)
            setattr(self, key, value)

    @property
    def filestore_keys(self):
        for desc in self.descriptors:
            for key, info in desc['data_keys'].items():
                try:
                    external = info['external']
                except KeyError:
                    continue

                try:
                    source, info = external.split(':', 1)
                except Exception:
                    pass
                else:
                    source = source.lower()
                    if source in ('filestore', ):
                        yield key

    @property
    def scan_id(self):
        return self.start_doc['scan_id']

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.header)

    def iterate_over_stream(self, stream_name, fill=False, key=None, **kwargs):
        if key is None:
            key = self.key

        if key:
            for event in self.header.events(fill=False,
                                            stream_name=stream_name):
                yield event['data'][key]

    def __iter__(self):
        yield from self.iterate_over_stream(self.stream, fill=False)


def combine_tables_on_time(header, names, *, method='ffill', **kwargs):
    '''Combine a fast-changing dataframe from databroker with one (or more)
    slow-changing ones, given its header.

    Parameters
    ----------
    header : Header
    names : list
        List of broker event stream names. The first will be taken as the
        primary (and in fact should probably be 'primary')
    method : {'ffill', 'bfill', 'nearest'}, optional
        Reindexing method
    **kwargs : dict, optional
        Passed to metadatastore.get_table

    Returns
    -------
    df : pd.DataFrame
    '''
    dfs = [header.table(stream_name=name, **kwargs)
           for name in names]

    primary_df = dfs[0]
    primary_index = primary_df.index

    try:
        times = primary_df['time']
    except KeyError:
        return dfs[0]

    dfs = ([primary_df] +
           [other.set_index('time').reindex(times, method=method)
            for other in dfs[1:]
            if 'time' in other])

    for df in dfs[1:]:
        df.index = primary_index
    return pd.concat(dfs, axis=1)


def get_combined_table(headers, name='primary',
                       combine_table_names=None,
                       **kwargs):
    # Functions the same as get_table, but also combines ('primary' and
    # 'motor2') when necessary (or whatever is in combine_table_names)

    if combine_table_names is None:
        if name == 'primary':
            combine_table_names = ['primary', 'motor2']
        else:
            combine_table_names = []

    if not combine_table_names:
        return [h.table(stream_name=name, **kwargs)
                for h in headers]

    try:
        headers.items()
    except AttributeError:
        pass
    else:
        headers = [headers]

    dfs = [combine_tables_on_time(header, names=combine_table_names, **kwargs)
           for header in headers]

    if dfs:
        return pd.concat(dfs)
    else:
        # edge case: no data
        return pd.DataFrame()



import logging

from ophyd import EpicsSignal
from ophyd.utils import DisconnectedError

logger = logging.getLogger(__name__)


class HxnScanStatus:
    '''Broadcast if a scan is in progress via PV

    Processed on every start/end document
    '''
    def __init__(self, running_pv):
        self._last_start = None
        self.running_signal = EpicsSignal(running_pv)
        self._running = False

    @property
    def running(self):
        return self._running

    @running.setter
    def running(self, running):
        self._running = running

        if running is None:
            running = 0

        try:
            self.running_signal.put(running)
        except DisconnectedError:
            logger.error('ScanRunning PV disconnected. Is the hxntools IOC running?')

    def __call__(self, name, doc):
        '''Bluesky callback with document info'''
        if name == 'start':
            self._last_start = doc
        if self._last_start and name in ('start', 'stop'):
            self.running = (name == 'start')



from ophyd import Device


def rename_device(dev, device_name, *,
                  special_attributes=None):
    '''Rename a device

    Parameters
    ----------
    dev : Device
    device_name : str
    special_attributes : str, optional
        If the attribute name matches any of these, the component instance
        will be named the same as the device.
        Defaults to ('readback', 'user_readback')
    '''
    dev.name = device_name

    if special_attributes is None:
        special_attributes = ('readback', 'user_readback')

    cls = dev.__class__
    for attribute in dev.component_names:
        obj = getattr(dev, attribute)
        component_cls = getattr(cls, attribute)

        if attribute in special_attributes:
            new_name = device_name
        else:
            new_name = component_cls.kwargs.get('name', attribute)
            new_name = '{}_{}'.format(device_name, new_name)

        obj.name = new_name

        if isinstance(obj, Device):
            rename_device(obj, new_name)


def rename_sub_devices(dev, *, special_attributes=None):
    '''Rename all sub-devices according to their keyword argument names'''
    cls = dev.__class__

    for attribute in dev.component_names:
        obj = getattr(dev, attribute)
        component_cls = getattr(cls, attribute)
        new_name = component_cls.kwargs.get('name', attribute)

        obj.name = new_name
        if isinstance(obj, Device):
            rename_device(obj, new_name,
                          special_attributes=special_attributes)


class NamedDevice(Device):
    '''
    NamedDevice will name components exactly as the 'name' argument
    specifies. Normally, it would be named based on the parent
    device's name: {parent_name}_{component_attribute_name}

    ''' + Device.__doc__

    def __init__(self, prefix, **kwargs):
        super().__init__(prefix, **kwargs)

        rename_sub_devices(self)



from __future__ import print_function

import logging

# imort XRF_DATA_KEY for back-compat
from databroker.assets.handlers import (Xspress3HDF5Handler,
                                        XS3_XRF_DATA_KEY as XRF_DATA_KEY)


logger = logging.getLogger(__name__)

FMT_ROI_KEY = 'entry/instrument/detector/NDAttributes/CHAN{}ROI{}'


def register(db):
    db.reg.register_handler(Xspress3HDF5Handler.HANDLER_NAME,
                            Xspress3HDF5Handler)



from __future__ import print_function
import logging
from pathlib import PurePath
from ophyd import (AreaDetector, CamBase, TIFFPlugin, Component as Cpt,
                   HDF5Plugin, Device, StatsPlugin, ProcessPlugin,
                   ROIPlugin, TransformPlugin)

from ophyd.areadetector.filestore_mixins import (
    FileStoreTIFF, FileStorePluginBase)

from .utils import makedirs
from .trigger_mixins import (HxnModalTrigger, FileStoreBulkReadable)


logger = logging.getLogger(__name__)


class MerlinTiffPlugin(TIFFPlugin, FileStoreBulkReadable, FileStoreTIFF,
                       Device):
    def mode_external(self):
        total_points = self.parent.mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

    def get_frames_per_point(self):
        mode = self.parent.mode_settings.mode.get()
        if mode == 'external':
            return 1
        else:
            return self.parent.cam.num_images.get()


class MerlinDetectorCam(CamBase):
    pass


class MerlinDetector(AreaDetector):
    cam = Cpt(MerlinDetectorCam, 'cam1:',
              read_attrs=[],
              configuration_attrs=['image_mode', 'trigger_mode',
                                   'acquire_time', 'acquire_period'],
              )


class MerlinFileStoreHDF5(FileStorePluginBase, FileStoreBulkReadable):
    _spec = 'TPX_HDF5'
    filestore_spec = _spec

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs.update([(self.file_template, '%s%s_%6.6d.h5'),
                                (self.file_write_mode, 'Stream'),
                                (self.compression, 'zlib'),
                                (self.capture, 1)
                                ])

    def stage(self):
        staged = super().stage()
        res_kwargs = {'frame_per_point': 1}
        logger.debug("Inserting resource with filename %s", self._fn)
        self._generate_resource(res_kwargs)

        return staged

    def make_filename(self):
        fn, read_path, write_path = super().make_filename()
        mode_settings = self.parent.mode_settings
        if mode_settings.make_directories.get():
            makedirs(read_path)
        return fn, read_path, write_path


class HDF5PluginWithFileStore(HDF5Plugin, MerlinFileStoreHDF5):
    def stage(self):
        mode_settings = self.parent.mode_settings
        total_points = mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

        # ensure that setting capture is the last thing that's done
        self.stage_sigs.move_to_end(self.capture)
        return super().stage()


class HxnMerlinDetector(HxnModalTrigger, MerlinDetector):
    hdf5 = Cpt(HDF5PluginWithFileStore, 'HDF1:',
               read_attrs=[],
               configuration_attrs=[],
               write_path_template='/data/%Y/%m/%d/',
               root='/data')

    proc1 = Cpt(ProcessPlugin, 'Proc1:')
    stats1 = Cpt(StatsPlugin, 'Stats1:')
    stats2 = Cpt(StatsPlugin, 'Stats2:')
    stats3 = Cpt(StatsPlugin, 'Stats3:')
    stats4 = Cpt(StatsPlugin, 'Stats4:')
    stats5 = Cpt(StatsPlugin, 'Stats5:')
    transform1 = Cpt(TransformPlugin, 'Trans1:')
    roi1 = Cpt(ROIPlugin, 'ROI1:')
    roi2 = Cpt(ROIPlugin, 'ROI2:')
    roi3 = Cpt(ROIPlugin, 'ROI3:')
    roi4 = Cpt(ROIPlugin, 'ROI4:')

    # tiff1 = Cpt(MerlinTiffPlugin, 'TIFF1:',
    #             read_attrs=[],
    #             configuration_attrs=[],
    #             write_path_template='/data/%Y/%m/%d/',
    #             root='/data')

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = ['hdf5', 'cam']
        if configuration_attrs is None:
            configuration_attrs = ['hdf5', 'cam']

        if 'hdf5' not in read_attrs:
            # ensure that hdf5 is still added, or data acquisition will fail
            read_attrs = list(read_attrs) + ['hdf5']

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

    def mode_internal(self):
        super().mode_internal()

        count_time = self.count_time.get()
        if count_time is not None:
            self.stage_sigs[self.cam.acquire_time] = count_time
            self.stage_sigs[self.cam.acquire_period] = count_time + 0.005

    def mode_external(self):
        super().mode_external()

        # NOTE: these values specify a debounce time for external triggering so
        #       they should be set to < 0.5 the expected exposure time, or at
        #       minimum the lowest possible dead time = 1.64ms
        expected_exposure = 0.001
        min_dead_time = 0.00164
        self.stage_sigs[self.cam.acquire_time] = expected_exposure
        self.stage_sigs[self.cam.acquire_period] = expected_exposure + min_dead_time

        self.cam.stage_sigs[self.cam.trigger_mode] = 'Trigger Enable'



from __future__ import print_function
import itertools
import logging

from ophyd import (AreaDetector, CamBase, TIFFPlugin, Component as Cpt,
                   HDF5Plugin, Device, StatsPlugin, ProcessPlugin,
                   ROIPlugin, EpicsSignal)
from ophyd.areadetector.plugins import PluginBase
from ophyd.areadetector import (EpicsSignalWithRBV as SignalWithRBV)
from ophyd.areadetector.filestore_mixins import (FileStoreTIFF,
                                                 FileStorePluginBase)

from .utils import (makedirs, make_filename_add_subdirectory)
from .trigger_mixins import (HxnModalTrigger, FileStoreBulkReadable)

from pathlib import PurePath

logger = logging.getLogger(__name__)


class DexelaTiffPlugin(TIFFPlugin, FileStoreBulkReadable, FileStoreTIFF,
                       Device):
    def mode_external(self):
        total_points = self.parent.mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

    def get_frames_per_point(self):
        mode = self.parent.mode_settings.mode.get()
        if mode == 'external':
            return 1
        else:
            return self.parent.cam.num_images.get()


class DexelaDetectorCam(CamBase):
    acquire_gain = Cpt(EpicsSignal, 'DEXAcquireGain')
    acquire_offset = Cpt(EpicsSignal, 'DEXAcquireOffset')
    binning_mode = Cpt(SignalWithRBV, 'DEXBinningMode')
    corrections_dir = Cpt(EpicsSignal, 'DEXCorrectionsDir', string=True)
    current_gain_frame = Cpt(EpicsSignal, 'DEXCurrentGainFrame')
    current_offset_frame = Cpt(EpicsSignal, 'DEXCurrentOffsetFrame')
    defect_map_available = Cpt(EpicsSignal, 'DEXDefectMapAvailable')
    defect_map_file = Cpt(EpicsSignal, 'DEXDefectMapFile', string=True)
    full_well_mode = Cpt(SignalWithRBV, 'DEXFullWellMode')
    gain_available = Cpt(EpicsSignal, 'DEXGainAvailable')
    gain_file = Cpt(EpicsSignal, 'DEXGainFile', string=True)
    load_defect_map_file = Cpt(EpicsSignal, 'DEXLoadDefectMapFile')
    load_gain_file = Cpt(EpicsSignal, 'DEXLoadGainFile')
    load_offset_file = Cpt(EpicsSignal, 'DEXLoadOffsetFile')
    num_gain_frames = Cpt(EpicsSignal, 'DEXNumGainFrames')
    num_offset_frames = Cpt(EpicsSignal, 'DEXNumOffsetFrames')
    offset_available = Cpt(EpicsSignal, 'DEXOffsetAvailable')
    offset_constant = Cpt(SignalWithRBV, 'DEXOffsetConstant')
    offset_file = Cpt(EpicsSignal, 'DEXOffsetFile', string=True)
    save_gain_file = Cpt(EpicsSignal, 'DEXSaveGainFile')
    save_offset_file = Cpt(EpicsSignal, 'DEXSaveOffsetFile')
    serial_number = Cpt(EpicsSignal, 'DEXSerialNumber')
    software_trigger = Cpt(EpicsSignal, 'DEXSoftwareTrigger')
    use_defect_map = Cpt(EpicsSignal, 'DEXUseDefectMap')
    use_gain = Cpt(EpicsSignal, 'DEXUseGain')
    use_offset = Cpt(EpicsSignal, 'DEXUseOffset')


class DexelaDetector(AreaDetector):
    cam = Cpt(DexelaDetectorCam, 'cam1:',
              read_attrs=[],
              configuration_attrs=['image_mode', 'trigger_mode',
                                   'acquire_time', 'acquire_period'],
              )


class DexelaFileStoreHDF5(FileStorePluginBase, FileStoreBulkReadable):
    _spec = 'TPX_HDF5'
    filestore_spec = _spec

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs.update([(self.file_template, '%s%s_%6.6d.h5'),
                                (self.file_write_mode, 'Stream'),
                                (self.compression, 'zlib'),
                                (self.capture, 1)
                                ])

    def stage(self):
        staged = super().stage()
        res_kwargs = {'frame_per_point': 1}
        logger.debug("Inserting resource with filename %s", self._fn)
        self._generate_resource(res_kwargs)

        return staged

    def make_filename(self):
        fn, read_path, write_path = super().make_filename()
        mode_settings = self.parent.mode_settings
        if mode_settings.make_directories.get():
            makedirs(read_path)
        return fn, read_path, write_path


class HDF5PluginWithFileStore(HDF5Plugin, DexelaFileStoreHDF5):
    def stage(self):
        mode_settings = self.parent.mode_settings
        total_points = mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

        # ensure that setting capture is the last thing that's done
        self.stage_sigs.move_to_end(self.capture)
        return super().stage()

    @property
    def write_path_template(self):
        path = self._write_path_template
        if not path.endswith('\\'):
            return f'{path}\\'
        return path

    @write_path_template.setter
    def write_path_template(self, value):
        self._write_path_template = value


class TransformPluginV2(PluginBase):
    '''TODO/NOTE: transform plugin from R2-4 wipes away all old records

    https://github.com/areaDetector/ADCore/commit/b39ddac4a4a8d14997b94052e45ce47d72ec3ae8
    '''
    _default_suffix = 'Trans1:'
    _suffix_re = 'Trans\d:'
    _html_docs = ['NDPluginTransform.html']
    _plugin_type = 'NDPluginTransform'

    type_ = Cpt(EpicsSignal, 'Type')


class HxnDexelaDetector(HxnModalTrigger, DexelaDetector):
    hdf5 = Cpt(HDF5PluginWithFileStore, 'HDF1:',
               read_attrs=[],
               configuration_attrs=[],
               write_path_template='Z:\\%Y\\%m\\%d',
               read_path_template='/data/%Y/%m/%d/',
               root='/data',
               path_semantics='windows')

    # tiff1 = Cpt(DexelaTiffPlugin, 'TIFF1:',
    #             read_attrs=[],
    #             configuration_attrs=[],
    #             write_path_template='Z:\\%Y\\%m\\%d',
    #             read_path_template='/data/%Y/%m/%d/',
    #             path_semantics='windows')

    proc1 = Cpt(ProcessPlugin, 'Proc1:')
    stats1 = Cpt(StatsPlugin, 'Stats1:')
    stats2 = Cpt(StatsPlugin, 'Stats2:')
    stats3 = Cpt(StatsPlugin, 'Stats3:')
    stats4 = Cpt(StatsPlugin, 'Stats4:')
    stats5 = Cpt(StatsPlugin, 'Stats5:')
    transform1 = Cpt(TransformPluginV2, 'Trans1:')
    roi1 = Cpt(ROIPlugin, 'ROI1:')
    roi2 = Cpt(ROIPlugin, 'ROI2:')
    roi3 = Cpt(ROIPlugin, 'ROI3:')
    roi4 = Cpt(ROIPlugin, 'ROI4:')

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = ['hdf5', 'cam']
        if configuration_attrs is None:
            configuration_attrs = ['hdf5', 'cam']

        if 'hdf5' not in read_attrs:
            # ensure that hdf5 is still added, or data acquisition will fail
            read_attrs = list(read_attrs) + ['hdf5']

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

    def mode_internal(self):
        super().mode_internal()

        count_time = self.count_time.get()
        if count_time is not None:
            self.stage_sigs[self.cam.acquire_time] = count_time
            self.stage_sigs[self.cam.acquire_period] = count_time + 0.005

        self.cam.stage_sigs[self.cam.trigger_mode] = 'Int. Software'

    def mode_external(self):
        super().mode_external()

        # NOTE: read-out time without binning is significantly higher than
        #       other detectors will drop frames unless dead_time is set
        #       properly on fly-scans. Consider eventually, however complicated
        #       it may be, to use the busy signals from the detectors that
        #       support it
        self.stage_sigs[self.cam.acquire_time] = 0.005
        self.stage_sigs[self.cam.acquire_period] = 0.060

        self.cam.stage_sigs[self.cam.trigger_mode] = 'Ext. Bulb'



from databroker.assets.handlers import HandlerBase, ImageStack


class HDF5DatasetSliceHandler(HandlerBase):
    """
    Handler for data stored in one Dataset of an HDF5 file.
    Parameters
    ----------
    filename : string
        path to HDF5 file
    key : string
        key of the single HDF5 Dataset used by this Handler
    frame_per_point : integer, optional
        number of frames to return as one datum, default 1
    swmr : bool, optional
        Open the hdf5 file in SWMR read mode. Only used when mode = 'r'.
        Default is False.
    """
    def __init__(self, filename, key, frame_per_point=1):
        self._fpp = frame_per_point
        self._filename = filename
        self._key = key
        self._file = None
        self._dataset = None
        self._data_objects = {}
        self.open()

    def get_file_list(self, datum_kwarg_gen):
        return [self._filename]

    def __call__(self, point_number):
        # Don't read out the dataset until it is requested for the first time.
        if not self._dataset:
            self._dataset = self._file[self._key]

        if point_number not in self._data_objects:
            start = point_number * self._fpp
            stop = (point_number + 1) * self._fpp
            self._data_objects[point_number] = ImageStack(self._dataset,
                                                          start, stop)
        return self._data_objects[point_number]

    def open(self):
        import h5py
        if self._file:
            return

        self._file = h5py.File(self._filename, 'r')

    def close(self):
        super(HDF5DatasetSliceHandler, self).close()
        self._file.close()
        self._file = None


class TimepixHDF5Handler(HDF5DatasetSliceHandler):
    """
    Handler for the 'AD_HDF5' spec used by Area Detectors.
    In this spec, the key (i.e., HDF5 dataset path) is always
    '/entry/detector/data'.
    Parameters
    ----------
    filename : string
        path to HDF5 file
    frame_per_point : integer, optional
        number of frames to return as one datum, default 1
    """
    _handler_name = 'TPX_HDF5'
    specs = {_handler_name}

    # TODO this is only different due to the hardcoded key being different?
    hardcoded_key = '/entry/instrument/detector/data'

    def __init__(self, filename, frame_per_point=1):
        super(TimepixHDF5Handler, self).__init__(
                filename=filename, key=self.hardcoded_key,
                frame_per_point=frame_per_point)


def register(db):
    db.reg.register_handler(TimepixHDF5Handler._handler_name,
                            TimepixHDF5Handler, overwrite=True)



from ophyd import (Component as Cpt)
from ophyd.mca import (MercuryDXP, EpicsMCARecord, EpicsDXPMultiElementSystem,
                       SoftDXPTrigger)


class HxnMercuryDetector(SoftDXPTrigger, EpicsDXPMultiElementSystem):
    '''DXP Mercury with 1 channel example'''
    dxp = Cpt(MercuryDXP, 'dxp1:')
    mca = Cpt(EpicsMCARecord, 'mca1')



import time
from ophyd import (Device, Component as Cpt)
from ophyd import (EpicsMotor, EpicsSignal)

import bluesky.plan_stubs as bps

anc350_dc_controllers = [2, 3, 4, 7]
# add 6 to this list if controller's moved back to the microscope rack
anc350_axis_counts = {1: 6,
                      2: 6,
                      3: 4,
                      4: 6,
                      5: 6,
                      6: 6,
                      7: 3,
                      # 8: 4,
                      }


class Anc350Axis(Device):
    motor = Cpt(EpicsMotor, 'Mtr')
    desc = Cpt(EpicsSignal, 'Mtr.DESC')
    frequency = Cpt(EpicsSignal, 'Freq-I', write_pv='Freq-SP')

    amplitude = Cpt(EpicsSignal, 'Ampl-I', write_pv='Ampl-SP')

    def __init__(self, prefix, *, axis_num=None, **kwargs):
        self.axis_num = int(axis_num)
        super(Anc350Axis, self).__init__(prefix, **kwargs)


class Anc350Controller(Device):
    dc_period = Cpt(EpicsSignal, 'DCPer-I', write_pv='DCPer-SP')
    dc_off_time = Cpt(EpicsSignal, 'DCOff-I', write_pv='DCOff-SP')
    dc_enable = Cpt(EpicsSignal, 'DC-I', write_pv='DC-Cmd')

    def __init__(self, prefix, **kwargs):
        super(Anc350Controller, self).__init__(prefix, **kwargs)

    def setup_dc(self, enable, period, off_time, verify=True):
        enable = 1 if enable else 0
        period = int(period)
        off_time = int(off_time)
        wait_group = 'anc_set_dc'

        yield from bps.abs_set(self.dc_period, period,
                               group=wait_group)
        yield from bps.abs_set(self.dc_off_time, off_time,
                               group=wait_group)
        yield from bps.abs_set(self.dc_enable, enable,
                               group=wait_group)
        if verify:
            yield from bps.wait(group=wait_group)


class HxnAnc350Axis(Anc350Axis):
    def __init__(self, controller, axis_num, **kwargs):
        prefix = 'XF:03IDC-ES{{ANC350:{}-Ax:{}}}'.format(controller, axis_num)
        name = f'anc_axis_{controller}'
        super().__init__(prefix, axis_num=axis_num, name=name, **kwargs)


class HxnAnc350Controller(Anc350Controller):
    def __init__(self, controller, **kwargs):
        prefix = 'XF:03IDC-ES{{ANC350:{}}}'.format(controller)
        name = f'anc_controller_{controller}'
        super().__init__(prefix, name=name, **kwargs)

        self.axes = {axis: HxnAnc350Axis(controller, axis)
                     for axis in range(anc350_axis_counts[controller])}


anc350_controllers = {controller: HxnAnc350Controller(controller)
                      for controller in anc350_axis_counts}


def _dc_status(controller, axis):
    pass


def _wait_tries(signal, value, tries=20, period=0.1):
    '''Wait up to `tries * period` for signal.get() to equal value'''

    # TODO set_and_wait?
    while tries > 0:
        tries -= 1
        if signal.get() == value:
            break

        time.sleep(period)


def _dc_toggle(axis, enable, freq, dc_period, off_time):
    print('Axis {} {}: '.format(axis.axis_num, axis.desc.value), end='')
    yield from bps.mov(axis.frequency, freq)
    print('frequency={}'.format(axis.frequency.get()))


def dc_toggle(enable, controllers=None, freq=100, dc_period=20, off_time=10):
    if controllers is None:
        controllers = anc350_dc_controllers

    for controller in controllers:
        print('Controller {}: '.format(controller), end='')
        controller = anc350_controllers[controller]

        try:
            yield from controller.setup_dc(enable, dc_period, off_time)
        except RuntimeError as ex:
            print('[Failed]', ex)
        except TimeoutError:
            print('Timed out - is the controller powered on?')
            continue
        else:
            if enable:
                print('Enabled duty cycling ({} off/{} on)'.format(
                      controller.dc_off_time.get(),
                      controller.dc_period.get()))
            else:
                print('Disabled duty cycling')

        for axis_num, axis in sorted(controller.axes.items()):
            print('\t', end='')
            yield from _dc_toggle(axis, enable, freq,
                                  dc_period, off_time)


def dc_on(*, frequency=100):
    yield from dc_toggle(True, freq=frequency)


def dc_off(*, frequency=1000):
    yield from dc_toggle(False, freq=frequency)



import sys
from os import path

from setuptools import find_packages, setup

import versioneer

# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 6)
if sys.version_info < min_version:
    error = """
hxntools does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(*(sys.version_info[:2] + min_version))
    sys.exit(error)

here = path.abspath(path.dirname(__file__))

description = "NSLS-II Hard X-ray Nanoprobe data acquisition tools"

readme_file = path.join(here, 'README.md')
long_description = description
if path.isfile(readme_file):
    with open(readme_file, encoding='utf-8') as f:
        long_description = f.read()

req_file = path.join(here, 'requirements.txt')
if path.isfile(req_file):
    with open(req_file) as f:
        # Parse requirements.txt, ignoring any commented-out lines.
        requirements = [line for line in f.read().splitlines()
                        if not line.startswith('#')]


setup(
    name="hxntools",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Brookhaven National Laboratory",
    author_email="DAMA@bnl.gov",
    url="https://github.com/NSLS-II-HXN/hxntools",
    python_requires='>={}'.format('.'.join(str(n) for n in min_version)),
    packages=find_packages(exclude=['docs', 'tests']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
)



from __future__ import print_function
import os
from ophyd import Signal


def makedirs(path, mode=0o777):
    '''Recursively make directories and set permissions'''
    # Permissions not working with os.makedirs -
    # See: http://stackoverflow.com/questions/5231901
    if not path or os.path.exists(path):
        return []

    head, tail = os.path.split(path)
    ret = makedirs(head, mode)
    try:
        os.mkdir(path)
    except OSError as ex:
        if 'File exists' not in str(ex):
            raise

    os.chmod(path, mode)
    ret.append(path)
    return ret


def ordered_dict_move_to_beginning(od, key):
    if key not in od:
        return

    value = od[key]
    items = list((k, v) for k, v in od.items()
                 if k != key)
    od.clear()
    od[key] = value
    od.update(items)


def make_filename_add_subdirectory(fn, read_path, write_path, *,
                                   make_directories=True,
                                   hash_characters=5):
    '''
    tag on a portion of the hash to reduce the number of files in one
    directory

    Parameters
    ----------
    fn : str
        Filename
    read_path : str
        Read path
    write_path : str
        Write path
    make_directories : bool, optional
        Make directories and set permissions (on the read_path)
    hash_characters : int, optional
        Number of characters to use from the hash
    '''
    hash_portion = fn[:hash_characters]
    read_path = os.path.join(read_path, hash_portion, '')
    write_path = os.path.join(write_path, hash_portion, '')

    if make_directories:
        makedirs(read_path)
    return fn, read_path, write_path



import logging

from pathlib import PurePath
from ophyd.areadetector.filestore_mixins import (FileStoreIterativeWrite,
                                                 FileStoreTIFF,
                                                 FileStorePluginBase,
                                                 )
from ophyd import (Device, Component as Cpt, AreaDetector, TIFFPlugin,
                   HDF5Plugin, StatsPlugin, ProcessPlugin, ROIPlugin,
                   TransformPlugin)
from ophyd import (EpicsSignal, EpicsSignalRO)
from ophyd.areadetector import (EpicsSignalWithRBV as SignalWithRBV, CamBase)
from .utils import makedirs
from .trigger_mixins import (HxnModalTrigger, FileStoreBulkReadable)


logger = logging.getLogger(__name__)


class ValueAndSets(Device):
    def __init__(self, prefix, *, read_attrs=None, **kwargs):
        if read_attrs is None:
            read_attrs = ['value']

        super().__init__(prefix, read_attrs=read_attrs, **kwargs)

    def get(self, **kwargs):
        return self.value.get(**kwargs)

    def put(self, value, **kwargs):
        if value:
            self.set_yes.put(1, **kwargs)
        else:
            self.set_no.put(1, **kwargs)


class TpxExtendedFrame(ValueAndSets):
    value = Cpt(SignalWithRBV, 'TPX_ExtendedFrame')
    set_no = Cpt(EpicsSignal, 'TPX_ExtendedFrameNo')
    set_yes = Cpt(EpicsSignal, 'TPX_ExtendedFrameYes')


class TpxSaveRaw(ValueAndSets):
    value = Cpt(SignalWithRBV, 'TPX_SaveToFile')
    set_no = Cpt(EpicsSignal, 'TPX_SaveToFileNo')
    set_yes = Cpt(EpicsSignal, 'TPX_SaveToFileYes')


class TimepixDetectorCam(CamBase):
    _html_docs = []

    tpx_corrections_dir = Cpt(EpicsSignal, 'TPXCorrectionsDir', string=True)
    tpx_dac = Cpt(EpicsSignalRO, 'TPXDAC_RBV')
    tpx_dac_available = Cpt(EpicsSignal, 'TPX_DACAvailable')
    tpx_dac_file = Cpt(EpicsSignal, 'TPX_DACFile', string=True)
    tpx_dev_ip = Cpt(SignalWithRBV, 'TPX_DevIp')

    tpx_frame_buff_index = Cpt(EpicsSignal, 'TPXFrameBuffIndex')
    tpx_hw_file = Cpt(EpicsSignal, 'TPX_HWFile', string=True)
    tpx_initialize = Cpt(SignalWithRBV, 'TPX_Initialize')
    tpx_load_dac_file = Cpt(EpicsSignal, 'TPXLoadDACFile')
    tpx_num_frame_buffers = Cpt(SignalWithRBV, 'TPXNumFrameBuffers')
    tpx_pix_config_file = Cpt(EpicsSignal, 'TPX_PixConfigFile', string=True)
    tpx_reset_detector = Cpt(EpicsSignal, 'TPX_resetDetector')

    tpx_raw_image_number = Cpt(EpicsSignal, 'TPXImageNumber')
    tpx_raw_prefix = Cpt(EpicsSignal, 'TPX_DataFilePrefix', string=True)
    tpx_raw_path = Cpt(EpicsSignal, 'TPX_DataSaveDirectory', string=True)

    tpx_start_sophy = Cpt(SignalWithRBV, 'TPX_StartSoPhy')
    tpx_status = Cpt(EpicsSignalRO, 'TPXStatus_RBV')
    tpx_sync_mode = Cpt(SignalWithRBV, 'TPXSyncMode')
    tpx_sync_time = Cpt(SignalWithRBV, 'TPXSyncTime')
    tpx_system_id = Cpt(EpicsSignal, 'TPXSystemID')
    tpx_trigger = Cpt(EpicsSignal, 'TPXTrigger')


class TimepixDetector(HxnModalTrigger, AreaDetector):
    _html_docs = []
    cam = Cpt(TimepixDetectorCam, 'cam1:',
              read_attrs=[],
              configuration_attrs=['tpx_corrections_dir', 'tpx_dac',
                                   'tpx_dac_file', 'tpx_dev_ip', 'tpx_hw_file',
                                   'tpx_system_id', 'tpx_pix_config_file',
                                   ])

    def mode_internal(self):
        super().mode_internal()

        count_time = self.count_time.get()
        if count_time is not None:
            self.cam.stage_sigs[self.cam.acquire_time] = count_time
            self.cam.stage_sigs[self.cam.acquire_period] = count_time + 0.020

    def mode_external(self):
        raise RuntimeError('Timepix external triggering not supported '
                           'reliably')


class TimepixTiffPlugin(TIFFPlugin, FileStoreBulkReadable, FileStoreTIFF,
                        Device):
    def mode_external(self):
        total_points = self.parent.mode_settings.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

    def get_frames_per_point(self):
        mode = self.parent.mode_settings.mode.get()
        if mode == 'external':
            return 1
        else:
            return self.parent.cam.num_images.get()


class TimepixFileStoreHDF5(FileStorePluginBase, FileStoreIterativeWrite):
    _spec = 'TPX_HDF5'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs.update([(self.file_template, '%s%s_%6.6d.h5'),
                                (self.file_write_mode, 'Stream'),
                                (self.compression, 'zlib'),
                                (self.capture, 1)
                                ])

    def stage(self):
        staged = super().stage()
        res_kwargs = {'frame_per_point': 1}
        logger.debug("Inserting resource with filename %s", self._fn)
        fn = PurePath(self._fn).relative_to(self.reg_root)
        self._resource = self._reg.register_resource(
            self._spec,
            str(self.reg_root), str(fn),
            res_kwargs)
        return staged

    def make_filename(self):
        fn, read_path, write_path = super().make_filename()
        if self.parent.make_directories.get():
            makedirs(read_path)
        return fn, read_path, write_path


class HDF5PluginWithFileStore(HDF5Plugin, TimepixFileStoreHDF5):
    def stage(self):
        total_points = self.parent.total_points.get()
        self.stage_sigs[self.num_capture] = total_points

        # ensure that setting capture is the last thing that's done
        self.stage_sigs.move_to_end(self.capture)
        return super().stage()


class HxnTimepixDetector(TimepixDetector):
    hdf5 = Cpt(HDF5PluginWithFileStore, 'HDF1:',
               read_attrs=[],
               configuration_attrs=[],
               write_path_template='/data/%Y/%m/%d/',
               root='/data')

    proc1 = Cpt(ProcessPlugin, 'Proc1:')
    stats1 = Cpt(StatsPlugin, 'Stats1:')
    stats2 = Cpt(StatsPlugin, 'Stats2:')
    stats3 = Cpt(StatsPlugin, 'Stats3:')
    stats4 = Cpt(StatsPlugin, 'Stats4:')
    stats5 = Cpt(StatsPlugin, 'Stats5:')
    transform1 = Cpt(TransformPlugin, 'Trans1:')
    roi1 = Cpt(ROIPlugin, 'ROI1:')
    roi2 = Cpt(ROIPlugin, 'ROI2:')
    roi3 = Cpt(ROIPlugin, 'ROI3:')
    roi4 = Cpt(ROIPlugin, 'ROI4:')

    # tiff1 = Cpt(TimepixTiffPlugin, 'TIFF1:',
    #             read_attrs=[],
    #             configuration_attrs=[],
    #             write_path_template='/data/%Y/%m/%d/',
    #             root='/data')

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = ['cam', 'hdf5']
        if configuration_attrs is None:
            configuration_attrs = ['cam', 'hdf5']

        if 'hdf5' not in read_attrs:
            # ensure that hdf5 is still added, or data acquisition will fail
            read_attrs = list(read_attrs) + ['hdf5']

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

        # signal aliases?
        self.total_points = self.mode_settings.total_points
        self.make_directories = self.mode_settings.make_directories

    def mode_internal(self):
        super().mode_internal()

        num_exposures = self.cam.num_exposures.get()
        cam = self.cam
        proc1 = self.proc1
        hdf5 = self.hdf5

        if num_exposures <= 1:
            cam.stage_sigs[cam.image_mode] = 'Single'
            hdf5.stage_sigs[hdf5.nd_array_port] = cam.port_name.get()
        else:
            # multiple exposures isn't supported by the timepix driver :(
            # use the process plugin to sum the images
            cam.stage_sigs[cam.image_mode] = 'Multiple'
            cam.stage_sigs[cam.num_images] = num_exposures
            hdf5.stage_sigs[hdf5.nd_array_port] = proc1.port_name.get()
            proc1.stage_sigs[proc1.num_filter] = num_exposures
            proc1.stage_sigs[proc1.reset_filter] = 'Yes'
            proc1.stage_sigs[proc1.auto_reset_filter] = 'Yes'
            proc1.stage_sigs[proc1.enable_filter] = 'Enable'
            proc1.stage_sigs[proc1.filter_type] = 'Sum'
            proc1.stage_sigs[proc1.filter_callbacks] = 'Array N only'

        proc1.stage_sigs[proc1.blocking_callbacks] = 'Yes'
        hdf5.stage_sigs[hdf5.blocking_callbacks] = 'Yes'



class HxnScanNumberPrinter:
    def __init__(self):
        self._last_start = None

    def __call__(self, name, doc):
        if name == 'start':
            self._last_start = doc
        if self._last_start is None:
            return
        if name in ('start', 'stop'):
            print('Scan ID: {scan_id} [{uid}]'.format(**self._last_start))



from __future__ import print_function
import logging

from ophyd import (EpicsSignalRO, DeviceStatus)
from ophyd import (Device, Component as Cpt)

logger = logging.getLogger(__name__)


class BeamStatusDetector(Device):
    shutter_status = Cpt(EpicsSignalRO, 'SR-EPS{PLC:1}Sts:MstrSh-Sts')
    beam_current = Cpt(EpicsSignalRO, 'SR:C03-BI{DCCT:1}I:Real-I')
    enabled = Cpt(EpicsSignalRO, 'SR:C03-EPS{PLC:1}Sts:ID_BE_Enbl-Sts')

    def __init__(self, prefix='', *, min_current=100.0, read_attrs=None,
                 configuration_attrs=None,
                 **kwargs):
        self._min_current = min_current

        if configuration_attrs is None:
            configuration_attrs = ['shutter_status', 'enabled', 'beam_current']
        if read_attrs is None:
            read_attrs = ['beam_current']

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

        self._shutter_ok = None
        self._current_ok = None
        self._enabled_ok = None
        self._last_status = None
        self._statuses = []

        self.shutter_status.subscribe(self._shutter_changed)
        self.beam_current.subscribe(self._current_changed)
        self.enabled.subscribe(self._enabled_changed)

    @property
    def min_current(self):
        '''The minimum current required to be considered usable'''
        return self._min_current

    @min_current.setter
    def min_current(self, value):
        '''The minimum current required to be considered usable'''
        self._min_current = value
        self._current_changed(value=self.sr_beam_current.get())

    def _shutter_changed(self, value=None, **kwargs):
        self._shutter_ok = (value == 1)
        self._check_status()

    def _current_changed(self, value=None, **kwargs):
        self._current_ok = (value > self._min_current)
        self._check_status()

    def _enabled_changed(self, value=None, **kwargs):
        self._enabled_ok = (value == 1)
        self._check_status()

    @property
    def status(self):
        return self._shutter_ok and self._current_ok and self._enabled_ok

    def _check_status(self):
        if None in (self._shutter_ok, self._current_ok, self._enabled_ok):
            return

        status = self.status

        if status:
            self._done()

        if status != self._last_status:
            logger.warning('Beam status changed:')

            if self._enabled_ok:
                logger.warning('Beamline is enabled')
            else:
                logger.warning('Beamline is NOT enabled')

            if self._shutter_ok:
                logger.warning('Shutters are open')
            else:
                logger.warning('Shutters are closed')

            if self._current_ok:
                logger.warning('Current meets threshold of %f',
                               self.min_current)
            else:
                logger.warning('Current does not meet threshold of %f',
                               self.min_current)

        self._last_status = status

    def _done(self):
        for status in self._statuses:
            status._finished()

    def trigger(self):
        status = DeviceStatus(self)

        if self.status:
            status._finished()
        else:
            self._statuses.append(status)

        if not status.done:
            logger.warning('---')
            logger.warning('Waiting for beam status to change...')
            logger.warning('---')

        return status

    def read(self):
        del self._statuses[:]
        return super().read()



from .timepix import (TimepixDetector, HxnTimepixDetector)
from .zebra import (HxnZebra, Zebra)
from .merlin import HxnMerlinDetector
from .beamstatus import BeamStatusDetector
from .trigger_mixins import (HxnModalBase, )
from .mercury import (HxnMercuryDetector, )
from .dexela import (HxnDexelaDetector, )
